/*

    File: ext2.c

    Copyright (C) 1998-2002  Christophe GRENIER <grenier@cgsecurity.org>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */
#include <stdio.h>
#include <string.h>
#include "types.h"
#include "common.h"
#include "ext2.h"
#include "intrface.h"
#include "fnctdsk.h"
#include "hdaccess.h"

#define EXT2_BLOCK_SIZE(s)      (EXT2_MIN_BLOCK_SIZE << (s)->s_log_block_size)

int check_EXT2(t_param_disk disk_car,t_diskext partition,const int debug)
{
  unsigned char buffer[8*0x200];
  if(hd_read_LBA(disk_car,8, &buffer, partition->lba)!=0)
  { return 1; }
  if(test_EXT2(disk_car,(struct ext2_super_block*)&buffer[0x400],partition,debug,0)!=0)
    return 1;
  set_EXT2_info(disk_car,(struct ext2_super_block*)&buffer[0x400],partition,debug,0);
  return 0;
}

int set_EXT2_info(t_param_disk disk_car, const struct ext2_super_block *sb,t_diskext partition,const int debug, const int dump_ind)
{
  partition->info[0]='\0';
  set_part_name(partition,sb->s_volume_name,16);
  /* sb->s_last_mounted seems to be unemployed in kernel 2.2.16 */
  if(EXT2_HAS_COMPAT_FEATURE(sb,EXT3_FEATURE_COMPAT_HAS_JOURNAL)!=0)
    strcat(partition->info,"EXT3 ");
  else
    strcat(partition->info,"EXT2 ");
  if(EXT2_HAS_RO_COMPAT_FEATURE(sb,EXT2_FEATURE_RO_COMPAT_LARGE_FILE)!=0)
    strcat(partition->info,"Large file ");
  if(EXT2_HAS_RO_COMPAT_FEATURE(sb,EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER)!=0)
    strcat(partition->info,"Sparse superblock ");
  if(EXT2_HAS_INCOMPAT_FEATURE(sb,EXT3_FEATURE_INCOMPAT_RECOVER)!=0)
    strcat(partition->info,"Recover ");
  if(EXT2_HAS_INCOMPAT_FEATURE(sb,EXT3_FEATURE_INCOMPAT_JOURNAL_DEV)!=0)
    strcat(partition->info,"Journal dev ");
  if(sb->s_block_group_nr!=0)
  {
    strcat(partition->info,"Backup superblock ");
    if(debug!=0)
    {
      ecrit_rapport("\nblock_group_nr %d\n",sb->s_block_group_nr);
    }
  }
  // last mounted => date
  return 0;
}

int recover_EXT2(t_param_disk disk_car, const struct ext2_super_block *sb,t_diskext partition,const int debug, const int dump_ind)
{
  if(test_EXT2(disk_car,sb,partition,debug,dump_ind)!=0)
    return 1;
  set_EXT2_info(disk_car,sb,partition,debug,dump_ind);
  partition->part_type=(unsigned char)P_LINUX;
  partition->part_size=sb->s_blocks_count*(EXT2_MIN_BLOCK<<sb->s_log_block_size);
  partition->boot_sector=sb->s_block_group_nr*sb->s_blocks_per_group;
  partition->blocksize=EXT2_MIN_BLOCK_SIZE<<sb->s_log_block_size;
  partition->lba-=sb->s_block_group_nr*sb->s_blocks_per_group*(EXT2_MIN_BLOCK<<sb->s_log_block_size);
  if(debug!=0)
  {
    ecrit_rapport("\nrecover_EXT2: s_block_group_nr=%u, s_blocks_per_group=%u\n",sb->s_block_group_nr,sb->s_blocks_per_group);
    ecrit_rapport("recover_EXT2: boot_sector=%lu, s_blocsize=%lu\n",partition->boot_sector,partition->blocksize);
    ecrit_rapport("recover_EXT2: s_blocks_count %u\n",sb->s_blocks_count);
    ecrit_rapport("recover_EXT2: part_size %lu\n",partition->part_size);
  }
return 0;
}

int test_EXT2(t_param_disk disk_car, const struct ext2_super_block *sb,t_diskext partition,const int debug, const int dump_ind)
{
  if(sb->s_magic==EXT2_SUPER_MAGIC)
  {
    if(dump_ind!=0)
    {
      ecrit_rapport("\nLinux magic value at %u/%u/%u\n", LBA2cylinder(disk_car,partition->lba),LBA2head(disk_car,partition->lba),LBA2sector(disk_car,partition->lba));
      /* There is a little offset ... */
      dump(stdscr,sb,0x200);
    }
    if (sb->s_free_blocks_count >= sb->s_blocks_count) return 1;
    if (sb->s_free_inodes_count >= sb->s_inodes_count) return 1;
    if (sb->s_errors!=0 &&
	(sb->s_errors != EXT2_ERRORS_CONTINUE) &&
	(sb->s_errors != EXT2_ERRORS_RO) &&
	(sb->s_errors != EXT2_ERRORS_PANIC))
      return 1;
    if ((sb->s_state & ~(EXT2_VALID_FS | EXT2_ERROR_FS))!=0)
      return 1;
    if (sb->s_blocks_count == 0) /* reject empty filesystem */
      return 1;
    if(sb->s_log_block_size>2)  /* block size max = 4096 */
      return 1;
    if(partition->part_size!=0 && (partition->part_size<sb->s_blocks_count*(EXT2_MIN_BLOCK<<sb->s_log_block_size)))
      return 1;
    if(EXT2_HAS_COMPAT_FEATURE(sb,EXT3_FEATURE_COMPAT_HAS_JOURNAL)!=0)
      partition->upart_type=UP_EXT3;
    else
      partition->upart_type=UP_EXT2;
    return 0;
  }
  return 1;
}
