/*

    File: dir_fat.c

    Copyright (C) 1998-2002  Christophe GRENIER <grenier@cgsecurity.org>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

#include <stdio.h>
#include <ctype.h>	/* toupper */
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "types.h"
#include "common.h"
#include "fat.h"
#include "lang.h"
#include "hdaccess.h"
#include "fnctdsk.h"
#include "testdisk.h"
#include "intrface.h"
#include "dir.h"
static const char *monstr[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
				"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

static void fat_write_display(WINDOW *window, const void*private_data);
static void fat_write_rapport(const void*private_data);
static int fat_is_directory(const void*private_data);
static unsigned long int fat_get_cluster(const void*private_data);

typedef struct fat_dir fat_file_data;
struct fat_dir
{
  wchar_t unicode[1000];
  time_t modtime;
  __u8    attr;
  __u32   start;		/* starting cluster number */
  __u32	size;			/* file size (in bytes) */
};

int date_dos2unix(unsigned short f_time,unsigned short f_date);
t_file_data *dir_fat_aux(char*buffer,int size);
static inline void fat16_towchar(wchar_t *dst, const __u8 *src, size_t len);
t_file_data *fat1x_rootdir(t_param_disk disk_car, const t_diskext partition, const int debug, struct fat_boot_sector*fat_header);
t_file_data *fat12_dir(t_param_disk disk_car, const t_diskext partition, t_dir_data *dir_data, const unsigned long int cluster);
t_file_data *fat16_dir(t_param_disk disk_car, const t_diskext partition, t_dir_data *dir_data, const unsigned long int cluster);
t_file_data *fat32_dir(t_param_disk disk_car, const t_diskext partition, t_dir_data *dir_data, const unsigned long int cluster);
int dump_fat_root_aux(WINDOW *window,t_param_disk disk_car, t_diskext partition, const int debug, struct fat_boot_sector*fat_header);


static inline void fat16_towchar(wchar_t *dst, const __u8 *src, size_t len)
{
	while (len--) {
		*dst++ = src[0] | (src[1] << 8);
		src += 2;
	}
}

t_file_data *dir_fat_aux(char*buffer,int size)
{
  struct msdos_dir_entry *de=(struct msdos_dir_entry*)buffer;
  wchar_t unicode[1000];
  unsigned char long_slots;
  t_file_data *dir_list=NULL;
  t_file_data *current_file=NULL;
GetNew:
  long_slots = 0;
  unicode[0]=0;
  if (de->name[0] == (__s8) DELETED_FLAG)
    goto RecEnd;
  if (de->attr == ATTR_EXT) {
    unsigned int i;
    struct msdos_dir_slot *ds;
    unsigned char id;
    unsigned char slot;
    unsigned char slots;
    unsigned char sum;
    unsigned char alias_checksum;
ParseLong:
    slots = 0;
    ds = (struct msdos_dir_slot *) de;
    id = ds->id;
    if (!(id & 0x40))
      goto RecEnd;
    slots = id & ~0x40;
    if (slots > 20 || !slots)	/* ceil(256 * 2 / 26) */
      goto RecEnd;
    long_slots = slots;
    alias_checksum = ds->alias_checksum;

    slot = slots;
    while (1) {
      int offset;

      slot--;
      offset = slot * 13;
      fat16_towchar(unicode + offset, ds->name0_4, 5);
      fat16_towchar(unicode + offset + 5, ds->name5_10, 6);
      fat16_towchar(unicode + offset + 11, ds->name11_12, 2);

      if (ds->id & 0x40) {
	unicode[offset + 13] = 0;
      }
      de++;
      if((void*)de>=(void*)(buffer+size))
	goto EODir;
      if (slot == 0)
	break;
      ds = (struct msdos_dir_slot *) de;
      if (ds->attr !=  ATTR_EXT)
	goto RecEnd;	/* XXX */
      if ((ds->id & ~0x40) != slot)
	goto ParseLong;
      if (ds->alias_checksum != alias_checksum)
	goto ParseLong;
    }
    if (de->name[0] == (__s8) DELETED_FLAG)
      goto RecEnd;
    if (de->attr ==  ATTR_EXT)
      goto ParseLong;
    if (IS_FREE(de->name) || (de->attr & ATTR_VOLUME))
      goto RecEnd;
    for (sum = 0, i = 0; i < 11; i++)
      sum = (((sum&1)<<7)|((sum&0xfe)>>1)) + de->name[i];
    if (sum != alias_checksum)
      long_slots = 0;
  }
RecEnd:
  if(!unicode[0] &&(de->attr != ATTR_EXT))
  { /* short name 8.3 */
    int i;
    int j=0;
    for(i=0;(i<8)&&(de->name[i]!=' ');i++)
      unicode[j++]=de->name[i];
    if(de->ext[0]!=' ')
    {
      unicode[j++]='.';
      for(i=0;(i<3)&&(de->ext[i]!=' ');i++)
	unicode[j++]=de->ext[i];
    }
    unicode[j]=0;
  }
  if (((de->attr != ATTR_EXT)||long_slots) && unicode[0] && ((__s8) unicode[0] != (__s8) DELETED_FLAG))
  {
    t_file_data *new_file=MALLOC(sizeof(*new_file));
    fat_file_data *new_fat_element=MALLOC(sizeof(*new_fat_element));
    memcpy(new_fat_element->unicode,unicode,sizeof(new_fat_element->unicode));
    new_fat_element->modtime=date_dos2unix(de->time,de->date);
    new_fat_element->attr=de->attr;
    new_fat_element->start=(de->starthi<<16)|de->start;
    new_fat_element->size=de->size;
    new_file->prev=current_file;
    new_file->next=NULL;
    new_file->file_data=new_fat_element;
//    ecrit_rapport("fat: new file\n");
//    fat_write_rapport(new_fat_element);
    if(current_file)
      current_file->next=new_file;
    else
      dir_list=new_file;
    current_file=new_file;
  }
  de++;
  if((void *)de<(void *)(buffer+size))
    goto GetNew;
EODir:
//  ecrit_rapport("dir_fat_aux dir_list %p\n",dir_list);
  return dir_list;
}

static void fat_write_rapport(const void*private_data)
{
  const fat_file_data *file=(const fat_file_data*)private_data;
  int i;
  static char datestr[80];
  struct tm *tm_p;
  tm_p = gmtime(&file->modtime);
  sprintf(datestr, "%2d-%s-%4d %02d:%02d",
      tm_p->tm_mday, monstr[tm_p->tm_mon],
      1900 + tm_p->tm_year, tm_p->tm_hour,
      tm_p->tm_min);
  if (file->attr & ATTR_VOLUME)
    ecrit_rapport("     <VOL>");
  else
    if (file->attr & ATTR_DIR)
      ecrit_rapport("     <DIR>");
    else
      ecrit_rapport("%10u",file->size);
  ecrit_rapport(" %s ",datestr);
//  ecrit_rapport("%8d ",file->start);
  for(i=0;file->unicode[i]&&(i<1000);i++)
    ecrit_rapport("%c", (char) file->unicode[i]);
  ecrit_rapport("\n");
}

static void fat_write_display(WINDOW *window, const void*private_data)
{
  const fat_file_data *file=(const fat_file_data*)private_data;
  int i;
  static char datestr[80];
  struct tm *tm_p;
  tm_p = gmtime(&file->modtime);
  sprintf(datestr, "%2d-%s-%4d %02d:%02d",
      tm_p->tm_mday, monstr[tm_p->tm_mon],
      1900 + tm_p->tm_year, tm_p->tm_hour,
      tm_p->tm_min);
  if (file->attr & ATTR_VOLUME)
    wdoprintf(window,"     <VOL>");
  else
    if (file->attr & ATTR_DIR)
      wdoprintf(window,"     <DIR>");
    else
      wdoprintf(window,"%10u",file->size);
  wdoprintf(window," %s ",datestr);
  for(i=0;file->unicode[i]&&(i<50);i++)
    wdoprintf(window,"%c", (char)file->unicode[i]);
  wdoprintf(window,"\n");
}

static int fat_is_directory(const void*private_data)
{
  const fat_file_data *file=(const fat_file_data*)private_data;
  return (file->attr & ATTR_DIR);
}

static unsigned long int fat_get_cluster(const void*private_data)
{
  const fat_file_data *file=(const fat_file_data*)private_data;
  return file->start;
}

static int day_n[] = { 0,31,59,90,120,151,181,212,243,273,304,334,0,0,0,0 };
		  /* JanFebMarApr May Jun Jul Aug Sep Oct Nov Dec */

/* Convert a MS-DOS time/date pair to a UNIX date (seconds since 1 1 70). */

int date_dos2unix(unsigned short f_time,unsigned short f_date)
{
	int month,year,secs;

	/* first subtract and mask after that... Otherwise, if
	   f_date == 0, bad things happen */
	month = ((f_date >> 5) - 1) & 15;
	year = f_date >> 9;
	secs = (f_time & 31)*2+60*((f_time >> 5) & 63)+(f_time >> 11)*3600+86400*
	    ((f_date & 31)-1+day_n[month]+(year/4)+year*365-((year & 3) == 0 &&
	    month < 2 ? 1 : 0)+3653);
			/* days since 1.1.70 plus 80's leap day */
	return secs;
}

t_file_data *fat12_dir(t_param_disk disk_car, const t_diskext partition, t_dir_data *dir_data, const unsigned long int first_cluster)
{
  struct fat_boot_sector*fat_header=(struct fat_boot_sector*)dir_data->private_dir_data;
  if(first_cluster==0)
    return fat1x_rootdir(disk_car,partition,dir_data->debug,fat_header);
  {
    t_file_data *dir_list;
    int cluster_size=fat_header->cluster_size;
    unsigned char *buffer_dir=MALLOC(0x200*cluster_size*5);
    unsigned int cluster;
    int nbr_cluster;
    memset(buffer_dir,0,0x200*cluster_size*5);
    for(cluster=first_cluster, nbr_cluster=0;
	(cluster!=(unsigned)0x0fff) && (cluster>=2) && (nbr_cluster<5);
	cluster=get_next_cluster(disk_car,partition, P_12FAT,fat_header->reserved, cluster), nbr_cluster++)
    {
      dword start=partition->lba+fat_header->reserved+fat_header->fats*fat_header->fat_length+(get_dir_entries(fat_header)*32+512-1)/512+(cluster-2)*cluster_size;
      if(dir_data->debug)
	ecrit_rapport("FAT12 cluster=%u(0x%x), pos=%lu\n",cluster,cluster,start);
      if(hd_read_LBA(disk_car, cluster_size, buffer_dir+0x200*cluster_size*nbr_cluster, start))
      {
	ecrit_rapport(m_ROOT_CLUSTER_RERR);
	FREE(buffer_dir);
	return NULL;
      }
    }
    dir_list=dir_fat_aux(buffer_dir,0x200*cluster_size*nbr_cluster);
    FREE(buffer_dir);
    return dir_list;
  }
}

t_file_data *fat16_dir(t_param_disk disk_car, const t_diskext partition, t_dir_data *dir_data, const unsigned long int first_cluster)
{
  struct fat_boot_sector*fat_header=(struct fat_boot_sector*)dir_data->private_dir_data;
  if(first_cluster==0)
    return fat1x_rootdir(disk_car,partition,dir_data->debug,fat_header);
  {
    t_file_data *dir_list=NULL;
    int cluster_size=fat_header->cluster_size;
    unsigned char *buffer_dir=MALLOC(0x200*cluster_size*5);
    unsigned int cluster;
    int nbr_cluster;
    memset(buffer_dir,0,0x200*cluster_size*5);
    /* Need to correct the test */
    for(cluster=first_cluster, nbr_cluster=0;
	((cluster&0xfff8)!=(unsigned)0xfff8) && (cluster>=2) && (nbr_cluster<5);
	cluster=get_next_cluster(disk_car,partition, P_16FAT,fat_header->reserved, cluster), nbr_cluster++)
    {
      dword start=partition->lba+fat_header->reserved+fat_header->fats*fat_header->fat_length+(get_dir_entries(fat_header)*32+512-1)/512+(cluster-2)*cluster_size;
      if(dir_data->debug)
	ecrit_rapport("FAT16 cluster=%u(0x%x), pos=%lu\n",cluster,cluster,start);
      if(hd_read_LBA(disk_car, cluster_size, buffer_dir+0x200*cluster_size*nbr_cluster, start))
      {
	ecrit_rapport(m_ROOT_CLUSTER_RERR);
	FREE(buffer_dir);
	return NULL;
      }
    }
    dir_list=dir_fat_aux(buffer_dir,0x200*cluster_size*nbr_cluster);
    FREE(buffer_dir);
    return dir_list;
  }
}

t_file_data *fat32_dir(t_param_disk disk_car, const t_diskext partition, t_dir_data *dir_data, const unsigned long int first_cluster)
{
  struct fat_boot_sector*fat_header=(struct fat_boot_sector*)dir_data->private_dir_data;
  if(!fat_header->cluster_size || !fat_header->root_cluster)
    return NULL;
  {
    t_file_data *dir_list;
    int cluster_size=fat_header->cluster_size;
    unsigned char *buffer_dir=MALLOC(0x200*cluster_size*5);
    unsigned int cluster;
    int nbr_cluster;
    memset(buffer_dir,0,0x200*cluster_size*5);
    /* Need to correct the test */
    for(cluster=first_cluster==0?fat_header->root_cluster:first_cluster, nbr_cluster=0;
	((cluster&0xffffff8)!=(unsigned)0xffffff8) && (cluster>=2) && (nbr_cluster<5);
	cluster=get_next_cluster(disk_car,partition, P_32FAT,fat_header->reserved, cluster), nbr_cluster++)
    {
      dword start=partition->lba+fat_header->reserved+fat_header->fats*fat_header->fat32_length+(cluster-2)*cluster_size;
      if(dir_data->debug)
	ecrit_rapport("FAT32 cluster=%u(0x%x), pos=%lu\n",cluster,cluster,start);
      if(hd_read_LBA(disk_car, cluster_size, buffer_dir+0x200*cluster_size*nbr_cluster, start))
      {
	ecrit_rapport(m_ROOT_CLUSTER_RERR);
	FREE(buffer_dir);
	return 0;
      }
    }
    dir_list=dir_fat_aux(buffer_dir,0x200*cluster_size*5);
    FREE(buffer_dir);
    return dir_list;
  }
}


t_file_data *fat1x_rootdir(t_param_disk disk_car, const t_diskext partition, const int debug, struct fat_boot_sector*fat_header)
{
  int root_size=((get_dir_entries(fat_header)-1)/16+1);
  dword start=0;
//  ecrit_rapport("fat1x_rootdir root_size=%d\n",root_size);
  {
    unsigned char buffer_dir[0x200*root_size];
    start=partition->lba+fat_header->reserved+fat_header->fats*fat_header->fat_length;
    if(hd_read_LBA(disk_car, root_size, &buffer_dir, start))
    {
      ecrit_rapport("FAT 1x");
      ecrit_rapport(m_ROOT_CLUSTER_RERR);
      return NULL;
    }
    return dir_fat_aux(buffer_dir,0x200*root_size);
  }
}

int dump_fat_root_aux(WINDOW *window,t_param_disk disk_car, t_diskext partition, const int debug, struct fat_boot_sector*fat_header)
{
  t_dir_data dir_data;
  dir_data.window=window;
  dir_data.write_rapport=fat_write_rapport;
  dir_data.write_display=fat_write_display;
  dir_data.is_directory=fat_is_directory;
  dir_data.get_cluster_inode=fat_get_cluster;
  dir_data.debug=debug;
  dir_data.private_dir_data=fat_header;
  switch(partition->upart_type)
  {
    case UP_FAT12:
      dir_data.get_dir=fat12_dir;
      break;
    case UP_FAT16:
      dir_data.get_dir=fat16_dir;
      break;
    case UP_FAT32:
      dir_data.get_dir=fat32_dir;
      break;
    default:
      return -1;
  }
  return dir_partition(disk_car,partition,&dir_data,0);
}

int dump_fat_root(WINDOW *window,t_param_disk disk_car, t_diskext partition, const int debug)
{
  t_sector buffer;
  if(hd_read_LBA(disk_car,1, &buffer, partition->lba+partition->boot_sector))
  {
    ecrit_rapport(m_CHKFAT_RERR); return 1;
  }
  dump_fat_root_aux(window,disk_car, partition, debug, (struct fat_boot_sector*)buffer);
  return 0;
}
